'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _require = require('mobx'),
    observable = _require.observable,
    action = _require.action;

var _require2 = require('./utils'),
    proxyGetters = _require2.proxyGetters,
    promiseTry = _require2.promiseTry;

/**
 * Wraps the given function in a task function.
 *
 * @param  {Function} fn
 * @return {Task}
 */


function createTask(fn, opts) {
  opts = _extends({
    swallow: false,
    state: 'pending'
  }, opts);

  // Track how many times this task was called.
  // Used to prevent premature state changes when the
  // task is called again before the first run completes.
  var callCount = 0;

  /**
   * The actual task function.
   */
  function task() {
    var _this = this,
        _arguments = arguments;

    var callCountWhenStarted = ++callCount;
    return promiseTry(function () {
      task.setState({ state: 'pending', error: undefined, result: undefined });
      return Promise.resolve(fn.apply(_this, _arguments)).then(function (result) {
        // If we called the task again before the first
        // one completes, we don't want to set to resolved before the last call completes.
        if (callCountWhenStarted === callCount) {
          task.setState({ state: 'resolved', error: undefined, result: result });
        }
        return result;
      });
    }).catch(function (err) {
      if (callCountWhenStarted === callCount) {
        task.setState({ state: 'rejected', error: err, result: undefined });
      }
      if (!opts.swallow) {
        throw err;
      }
    });
  }

  var taskStateSchema = {
    state: opts.state,
    error: opts.error,
    result: opts.result,
    get pending() {
      return taskState.state === 'pending';
    },
    get resolved() {
      return taskState.state === 'resolved';
    },
    get rejected() {
      return taskState.state === 'rejected';
    }
  };

  var taskStateSchemaKeys = Object.keys(taskStateSchema);
  var taskState = observable.object(taskStateSchema, {
    error: observable.ref
  });

  setupTask(task, taskState, taskStateSchemaKeys, opts);
  return task;
}

/**
 * Assigns the task methods and state to the given function.
 */
function setupTask(fn, taskState, taskStateSchemaKeys, opts) {
  var setup = function setup(func) {
    return setupTask(func, taskState, taskStateSchemaKeys, opts);
  };
  proxyGetters(fn, taskState, taskStateSchemaKeys);
  Object.assign(fn, {
    /**
     * Patch `bind` so we always return
     * the task function (with the additional properties)
     */
    bind: function bind() {
      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      var bound = Function.prototype.bind.apply(fn, args);
      return setup(bound);
    },
    /**
     * Wraps the task and returns the new function with the task stuff attached.
     *
     * @param {Function} wrapper
     * Invoked with the inner function as the only parameter.
     *
     * @return {Task}
     * Wrapped function as a task.
     */
    wrap: function wrap(wrapper) {
      return setup(wrapper(function wrapped() {
        return fn.apply(this, arguments);
      }));
    },
    /**
     * Assigns the given properties to the task.
     * E.g. task.setState({ state: 'resolved', result: 1337 })
     */
    setState: action(function (opts) {
      Object.assign(taskState, opts);
      return fn;
    }),
    /**
     * Given an object, returns the value for the key which equals the
     * current state, or undefined if not specified.
     */
    match: function match(obj) {
      var state = taskState.state;
      var match = obj[state];

      if (!match) {
        return undefined;
      }

      if (state === 'resolved') {
        return match(taskState.result);
      }

      if (state === 'rejected') {
        return match(taskState.error);
      }

      return match();
    },
    /**
     * Resets the state to what it was when initialized.
     */
    reset: function reset() {
      fn.setState({
        state: opts.state,
        result: opts.result,
        error: opts.error
      });
      return fn;
    }
  });
  return fn;
}

/**
 * Returns a function, which returns either a decorator, a task, or a decorator factory.
 */
function taskCreatorFactory(opts) {
  /**
   * Decorator to make async functions "so fucking graceful", by maintaining observables for errors
   * and running state.
   */
  return function task(arg1, arg2, arg3) {
    if (typeof arg1 === 'function') {
      // regular invocation
      return createTask(arg1, _extends({}, opts, arg2));
    }

    var makeDecorator = function makeDecorator(innerOpts) {
      return function decorator(target, name, descriptor) {
        var get = descriptor.get;
        if (descriptor.value) {
          var fn = descriptor.value;
          delete descriptor.writable;
          delete descriptor.value;
          get = function get() {
            return fn;
          };
        }

        // In IE11 calling Object.defineProperty has a side-effect of evaluating the
        // getter for the property which is being replaced. This causes infinite
        // recursion and an "Out of stack space" error.
        // Credit: autobind-decorator source
        var definingProperty = false;
        return {
          get: function getter() {
            /* istanbul ignore next */
            if (definingProperty) {
              return get.apply(this);
            }

            var fn = get.apply(this, arguments);
            var wrapped = createTask(fn, _extends({}, opts, innerOpts));
            definingProperty = true;
            Object.defineProperty(this, name, {
              value: wrapped,
              configurable: true,
              writable: true
            });
            definingProperty = false;
            return wrapped;
          },
          set: function setter(newValue) {
            Object.defineProperty(this, name, {
              configurable: true,
              writable: true,
              // IS enumerable when reassigned by the outside word
              enumerable: true,
              value: newValue
            });

            return newValue;
          }
        };
      };
    };

    // decorator invocation
    if (typeof arg2 === 'string') {
      // parameterless - @task method()
      return makeDecorator()(arg1, arg2, arg3);
    }

    // parameters - @task({ state: 'resolved' }) method()
    return makeDecorator(_extends({}, opts, arg1));
  };
}

var task = taskCreatorFactory();
task.resolved = taskCreatorFactory({ state: 'resolved' });
task.rejected = taskCreatorFactory({ state: 'rejected' });

module.exports = task;
//# sourceMappingURL=task.js.map